"""
Michael Hirsch
Python wrapper of the venerable LOWTRAN7 atmospheric absorption and solar transmission
model circa 1992.

Note: specified Lowtran7 model limitations include
wlcminvstep >= 20 cm^-1
0 <= wlcminv <= 50000

historical note:
developed on CDC CYBER, currently runs on 32-bit single float--this can cause loss of numerical
precision, future would like to ensure full LOWTRAN7 code can run at 64-bit double float.

user manual:
www.dtic.mil/dtic/tr/fulltext/u2/a206773.pdf

"""
from pathlib import Path
from pandas import read_csv
from dateutil.parser import parse
import logging
import xarray
import numpy as np
#
try:
    import lowtran7   #don't use dot in front, it's linking to .dll, .pyd, or .so
except ImportError as e:
    raise ImportError('you must compile the Fortran code first. f2py -m lowtran7 -c lowtran7.f {}'.format(e))


def loopuserdef(c1:dict) -> xarray.DataArray:
    """
    golowtran() is for scalar parameters only
    (besides vector of wavelength, which Lowtran internally loops over)

    wmol, p, t must all be vector(s) of same length
    """

    wmol = np.atleast_2d(c1['wmol'])
    P = np.atleast_1d(c1['p'])
    T = np.atleast_1d(c1['t'])
    time = np.atleast_1d(c1['time'])

    assert wmol.shape[0] == len(P) == len(T) == len(time),'WMOL, P, T,time must be vectors of equal length'

    N = len(P)
# %% 3-D array indexed by metadata
    TR = xarray.Dataset(coords={'time':time, 'wavelength_nm':None, 'angle_deg':None})

    for i in range(N):
        c = c1.copy()
        c['wmol'] = wmol[i,:]
        c['p'] = P[i]
        c['t'] = T[i]
        c['time'] = time[i]

        TR = TR.merge(_golowtran(c))

 #   TR = TR.sort_index(axis=0) # put times in order, sometimes CSV is not monotonic in time.

    return TR


def loopangle(c1:dict) -> xarray.Dataset:
    """
    loop over "ANGLE"
    """
    angles = np.atleast_1d(c1['angle'])
    TR = xarray.Dataset(coords={'wavelength_nm':None,'angle_deg':angles})

    for a in angles:
        c = c1.copy()
        c['angle'] = a
        TR = TR.merge(_golowtran(c))

    return TR


def _golowtran(c1:dict) -> xarray.Dataset:
    """directly run Fortran code"""
# %% default parameters
    if 'time' not in c1:
        c1['time'] = None

    defp = ('h1','h2','angle','im','iseasn','ird1','range_km','zmdl','p','t')
    for p in defp:
        if p not in c1:
            c1[p] =  0

    if 'wmol' not in c1:
        c1['wmol']=[0]*12
# %% input check
    assert len(c1['wmol']) == 12,'see Lowtran user manual for 12 values of WMOL'
    assert np.isfinite(c1['h1']),'per Lowtran user manual Table 14, H1 must always be defined'
# %% setup wavelength
    wlcminv,wlcminvstep,nwl = nm2lt7(c1['wlnmlim'])
    if wlcminvstep<5:
        logging.critical('minimum resolution 5 cm^-1, specified resolution 20 cm^-1')
    if not ((0<=wlcminv) & (wlcminv<=50000)).all():
       logging.critical('specified model range 0 <= wlcminv <= 50000')
# %% invoke lowtran
    """
    Note we invoke case "3a" from table 14, only observer altitude and apparent
    angle are specified
    """
    Tx,V,Alam,trace,unif,suma,irrad,sumvv = lowtran7.lwtrn7(
                            True, nwl, wlcminv[1], wlcminv[0], wlcminvstep,
                            c1['model'], c1['itype'], c1['iemsct'], c1['im'],
                            c1['iseasn'], c1['ird1'],
                            c1['zmdl'], c1['p'], c1['t'], c1['wmol'],
                            c1['h1'], c1['h2'], c1['angle'], c1['range_km'])

    dims = ('time','wavelength_nm','angle_deg')
    TR = xarray.Dataset({'transmission':(dims,Tx[:,9][None,:,None]),
                         'radiance':(dims,sumvv[None,:,None]),
                         'irradiance':(dims,irrad[:,0][None,:,None]),
                         'pathscatter':(dims,irrad[:,2][None,:,None]) },
                        coords={'time':[c1['time']],
                                'wavelength_nm':Alam*1e3,
                                'angle_deg':[c1['angle']]})

    return TR


def nm2lt7(wlnm):
    """converts wavelength in nm to cm^-1"""
    wlcminvstep = 5 # minimum meaningful step is 20, but 5 is minimum before crashing lowtran
    wlnm= np.asarray(wlnm,dtype=float) #for proper division
    wlcminv = 1.e7/wlnm
    nwl = int(np.ceil((wlcminv[0]-wlcminv[1])/wlcminvstep))+1 #yes, ceil

    return wlcminv,wlcminvstep,nwl


def scatter(c1:dict) -> xarray.Dataset:
# %% low-level Lowtran configuration for this scenario, don't change
    c1.update({
        'itype':  3,  # 3: observer to space
        'iemsct': 2,  #  2: radiance model
        })
# %% TR is 3-D array with axes: time, wavelength, and [transmission,radiance]
    return loopangle(c1)


def irradiance(c1:dict) -> xarray.Dataset:
    c1.update({
        'itype':3,   # 3: observer to space
        'iemsct':3,  # 3: solar irradiance
         })

    return loopangle(c1)


def radiance(c1:dict) -> xarray.Dataset:

    c1.update({
        'itype':  3,  # 3: observer to space
        'iemsct': 1,  # 1: thermal radiance model  2: radiance model
        })
#%% TR is 3-D array with axes: time, wavelength, and [transmission,radiance]
    return loopangle(c1)


def transmittance(c1:dict) -> xarray.Dataset:

    c1.update({
        'itype':3,   # 3: observer to space
        'iemsct':0,  # 0: transmittance
        })

    return loopangle(c1)


def horizrad(infn:Path, outfn:Path, c1:dict) -> xarray.Dataset:
    """
    read CSV, simulate, write, plot
    """
    if infn is not None:
        infn = Path(infn).expanduser()

        if infn.suffix=='.h5':
            TR = xarray.open_dataset(infn)
            return TR

    c1.update({'model':0,  # 0: user meterological data
        'itype':1,  # 1: horizontal path
        'iemsct':1, # 1: radiance model
        'im': 1,    # 1: for horizontal path (see Lowtran manual p.42)
        'ird1': 1,  # 1: use card 2C2)
        })
#%% read csv file
    if not infn: # demo mode
        c1['p']=[949., 959.]
        c1['t']=[283.8, 285.]
        c1['wmol']=[[93.96,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.],
                    [93.96,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.]]
        c1['time']=[parse('2017-04-05T12'),
                    parse('2017-04-05T18')]
    else: # read csv, normal case
        PTdata = read_csv(infn)
        c1['p'] = PTdata['p']
        c1['t'] = PTdata['Ta']
        c1['wmol'] = np.zeros((PTdata.shape[0], 12))
        c1['wmol'][:,0] = PTdata['RH']
        c1['time'] = [parse(t) for t in PTdata['time']]
#%% TR is 3-D array with axes: time, wavelength, and [transmission,radiance]
    TR = loopuserdef(c1)

    return TR


def horiztrans(c1:dict) -> xarray.Dataset:

    c1.update({'model':5, # 5: subartic winter
        'itype':1, # 1: horizontal path
        'iemsct':0, # 0: transmittance model
        'im': 0, # 1: for user-defined atmosphere on horizontal path (see Lowtran manual p.42)
        'ird1': 0, # 1: use card 2C2
        })

#    c1.update({'model':0, # 0: user meterological data
#        'itype':1, # 1: horizontal path
#        'iemsct':0, # 0: transmittance model
#        'im': 1, # 1: for user-defined atmosphere on horizontal path (see Lowtran manual p.42)
#        'ird1': 1, # 1: use card 2C2
#        'p':949.,
#        't':283.8,
#        'wmol':[93.96,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.]
#        })


    return _golowtran(c1)



#!/usr/bin/env python
"""
Horizontal cases are for special use by advanced users.

lowtran manual p.36 specify height H1 and RANGE
"""
from matplotlib.pyplot import show
#
import lowtran
from lowtran.plots import plothoriz

if __name__=='__main__':

    from argparse import ArgumentParser
    p = ArgumentParser(description='Lowtran 7 interface')
    p.add_argument('-z','--obsalt',help='altitude of bother observers on horizontal path [km]',type=float,default=0.3)
    p.add_argument('-r','--range_km',help='range between observers on horizontal path [km]', type=float, default=1.0)
    p.add_argument('-a','--zenang',help='zenith angle [deg]  can be single value or list of values',type=float,default=0.)
    p.add_argument('-w','--wavelen',help='wavelength range nm (start,stop)',type=float,nargs=2,default=(200,30000))
    p=p.parse_args()

    c1={'zmdl':p.obsalt,
        'h1': p.obsalt,
        'range_km':p.range_km,
        'wlnmlim': p.wavelen,
        }

    TR = lowtran.horiztrans(c1).squeeze()

    plothoriz(TR, c1)

    show()



#!/usr/bin/env python
"""
For Irradiance, the zenith angle is locked to the zenith angle of the sun.
Implicitly, your sensor is looking at the sun and that's the only choice per
Lowtran manual p. 36 s3.2.3.1
"""
from matplotlib.pyplot import show
#
import lowtran
from lowtran.plots import plotirrad

if __name__=='__main__':

    from argparse import ArgumentParser
    p = ArgumentParser(description='Lowtran 7 interface')
    p.add_argument('-z','--obsalt',help='altitude of observer [km]', type=float, default=0.)
    p.add_argument('-a','--zenang',help='zenith angle [deg]  of sun or moon',nargs='+',type=float,default=[0,60,80])
    p.add_argument('-w','--wavelen',help='wavelength range nm (start,stop)',type=float,nargs=2,default=(200,25000))
    p.add_argument('--model',help='0-6, see Card1 "model" reference. 5=subarctic winter',type=int,default=5)
    p=p.parse_args()

    c1={'model':p.model,
        'h1': p.obsalt,
        'angle': p.zenang, # zenith angle of sun or moon
        'wlnmlim': p.wavelen,
        }

    I = lowtran.irradiance(c1)

    plotirrad(I, c1, True)

    show()



#!/usr/bin/env python
from matplotlib.pyplot import show
#
import lowtran
from lowtran.plots import plottrans

if __name__=='__main__':

    from argparse import ArgumentParser
    p = ArgumentParser(description='Lowtran 7 interface')
    p.add_argument('-z','--obsalt',help='altitude of observer [km]',type=float,default=0.)
    p.add_argument('-a','--zenang',help='observer zenith angle [deg]',type=float,nargs='+',default=[0,60,80])
    p.add_argument('-w','--wavelen',help='wavelength range nm (start,stop)',type=float,nargs=2,default=(200,30000))
    p.add_argument('--model',help='0-6, see Card1 "model" reference. 5=subarctic winter',type=int,default=5)
    p=p.parse_args()

    c1={'model':p.model,
        'h1': p.obsalt,
        'angle': p.zenang,
        'wlnmlim': p.wavelen,
        }

    TR = lowtran.transmittance(c1)

    plottrans(TR, c1)

    show()



import numpy as np
import xarray
from matplotlib.pyplot import figure
#
h = 6.62607004e-34
c = 299792458
UNITS='ster$^{-1}$ cm$^{-2}$ $\mu$m$^{-1}$]'
plotNp = False

def plotscatter(irrad:xarray.Dataset, c1:dict, log:bool=False):

    fg = figure()
    axs = fg.subplots(2, 1, sharex=True)

    transtxt = 'Transmittance'

    ax = axs[0]
    ax.plot(irrad.wavelength_nm, irrad['transmission'].squeeze())
    ax.set_title(transtxt)
    ax.set_ylabel('Transmission (unitless)')
    ax.grid(True)
    ax.legend(irrad.angle_deg.values)

    ax = axs[1]
    if plotNp:
        Np = (irrad['pathscatter']*10000) * (irrad.wavelength_nm*1e9)/(h*c)
        ax.plot(irrad.wavelength_nm, Np)
        ax.set_ylabel('Photons [s$^{-1}$ '+UNITS)
    else:
        ax.plot(irrad.wavelength_nm, irrad['pathscatter'].squeeze())
        ax.set_ylabel('Radiance [W '+UNITS)

    ax.set_xlabel('wavelength [nm]')
    ax.set_title('Single-scatter Path Radiance')
    ax.invert_xaxis()
    ax.autoscale(True,axis='x',tight=True)
    ax.grid(True)


    if log:
        ax.set_yscale('log')
#        ax.set_ylim(1e-8,1)

    try:
        fg.suptitle(f'Obs. to Space: zenith angle: {c1["angle"]} deg., ')
        #{datetime.utcfromtimestamp(irrad.time.item()/1e9)}
    except (AttributeError,TypeError):
        pass


def plotradiance(irrad:xarray.Dataset, c1:dict, log:bool=False):
    fg = figure()
    axs = fg.subplots(2, 1, sharex=True)

    transtxt = 'Transmittance Observer to Space'

    ax = axs[0]
    ax.plot(irrad.wavelength_nm, irrad['transmission'].squeeze())
    ax.set_title(transtxt)
    ax.set_ylabel('Transmission (unitless)')
    ax.grid(True)

    ax = axs[1]
    if plotNp:
        Np = (irrad['radiance']*10000) * (irrad.wavelength_nm*1e9)/(h*c)
        ax.plot(irrad.wavelength_nm, Np)
        ax.set_ylabel('Photons [s$^{-1}$ '+UNITS)
    else:
        ax.plot(irrad.wavelength_nm, irrad['radiance'].squeeze())
        ax.set_ylabel('Radiance [W '+UNITS)

    ax.set_xlabel('wavelength [nm]')
    ax.set_title('Atmospheric Radiance')
    ax.invert_xaxis()
    ax.autoscale(True,axis='x',tight=True)
    ax.grid(True)

    if log:
        ax.set_yscale('log')
        ax.set_ylim(1e-8,1)

    try:
        fg.suptitle(f'Obs. zenith angle: {c1["angle"]} deg., ')
        #{datetime.utcfromtimestamp(irrad.time.item()/1e9)}
    except (AttributeError,TypeError):
        pass

def plotradtime(TR:xarray.Dataset, c1:dict, log:bool=False):
    """
    make one plot per time for now.

    TR: 3-D array: time, wavelength, [transmittance, radiance]

    radiance is currently single-scatter solar
    """

    for t in TR.time: # for each time
        plotirrad(TR.sel(time=t), c1, log)


def plottrans(T:xarray.Dataset, c1:dict, log:bool=False):
    ax = figure().gca()

    h = ax.plot(T.wavelength_nm, T['transmission'].squeeze())

    ax.set_xlabel('wavelength [nm]')
    ax.set_ylabel('transmission (unitless)')
    ax.set_title(f'Transmittance Ground-Space: Obs. zenith angle {c1["angle"]} deg.')
    #ax.legend(loc='best')
    ax.grid(True)
    if log:
        ax.set_yscale('log')
        ax.set_ylim(1e-5,1)
    else:
        ax.set_ylim(0,1)
    ax.invert_xaxis()
    ax.autoscale(True,axis='x',tight=True)
    ax.legend(h, T.angle_deg.values)


def plotirrad(irrad:xarray.Dataset, c1:dict, log:bool=False):
    fg = figure()
    axs = fg.subplots(2,1,sharex=True)

#    if c1['isourc'] == 0:
    stxt = "Sun's"
#    elif c1['isourc'] == 1:
#        stxt = "Moon's"
#    else:
#        raise ValueError(f'ISOURC={c1["isourc"]} not defined case')

    stxt += f' zenith angle {irrad.angle_deg.values} deg., Obs. height {c1["h1"]} km. '
    try:
        stxt += np.datetime_as_string(irrad.time)[:-10]
    except (AttributeError,TypeError):
        pass

    fg.suptitle(stxt)

    if c1['iemsct'] == 3:
        key='irradiance'
        transtxt = 'Transmittance Observer to Space'
    elif c1['iemsct'] == 1:
        key='radiance'
        transtxt = 'Transmittance Observer to Observer'

    #irrad.['transmission'].plot()

    ax = axs[0]
    h = ax.plot(irrad.wavelength_nm, irrad['transmission'].squeeze())
    ax.set_title(transtxt)
    ax.set_ylabel('Transmission (unitless)')
    ax.grid(True)
    try:
        ax.legend(h, irrad.angle_deg.values)
    except AttributeError:
        pass

    ax = axs[1]
    ax.plot(irrad.wavelength_nm, irrad[key].squeeze())
    ax.set_xlabel('wavelength [nm]')
    ax.invert_xaxis()
    ax.grid(True)

    if c1['iemsct'] == 3:
        ttxt= 'Irradiance '
        ax.set_ylabel('Solar Irradiance [W '+UNITS)
        ax.set_title(ttxt)
    elif c1['iemsct'] ==1:
        ttxt = 'Single-scatter Radiance '
        ax.set_ylabel('Radiance [W '+UNITS)
        ax.set_title(ttxt)

    if log:
        ax.set_yscale('log')
        ax.set_ylim(1e-8,1)

    ax.autoscale(True,axis='x',tight=True)


def plothoriz(trans:xarray.Dataset, c1:dict, log:bool=False):
    ax = figure().gca()

    ax.plot(trans.wavelength_nm, trans['transmission'].squeeze())

    ax.set_xlabel('wavelength [nm]')
    ax.set_ylabel('transmission (unitless)')
    ax.set_title(f'Transmittance Horizontal \n {c1["range_km"]} km path @ {c1["h1"]} km altitude')
    #ax.legend(loc='best')
    ax.grid(True)
    if log:
        ax.set_yscale('log')
        ax.set_ylim(1e-5,1)
    else:
        ax.set_ylim(0,1)
    ax.invert_xaxis()
    ax.autoscale(True,axis='x',tight=True)



#!/usr/bin/env python
"""
Horizontal cases are for special use by advanced users.

"""
from pathlib import Path
from matplotlib.pyplot import show
#
import lowtran
from lowtran.plots import plotradtime

if __name__=='__main__':

    from argparse import ArgumentParser
    p = ArgumentParser(description='Lowtran 7 interface')
    p.add_argument('ptfn',help='csv file with time,relative humidity [%],ambient temperature [K], total pressure (millibar)',nargs='?')
    p.add_argument('-z','--obsalt',help='altitude of observer [km]',type=float,default=0.05)
    p.add_argument('-w','--wavelen',help='wavelength range nm (start,stop)',type=float,nargs=2,default=(200,30000))
    p.add_argument('-o','--outfn',help='HDF5 file to write')
    p=p.parse_args()

# %% low-level Lowtran configuration for this scenario, don't change
    c1={'range_km':p.obsalt,
        'zmdl':p.obsalt,
        'h1':p.obsalt,
        'wlnmlim': p.wavelen,
        }

    TR = lowtran.horizrad(p.ptfn,p.outfn,c1)

# %% write to HDF5
    if p.outfn:
        outfn = Path(p.outfn).expanduser()
        print('writing', outfn)
        TR.to_netcdf(outfn)

    plotradtime(TR, c1)

    show()



#!/usr/bin/env python
install_requires=['python-dateutil','numpy','xarray']
tests_require=['pytest','nose','coveralls']
# %%
from setuptools import find_packages
from numpy.distutils.core import setup,Extension

ext = [Extension(name='lowtran7',
                 sources=['lowtran7.f'],
                 f2py_options=['--quiet'])]

setup(name='lowtran',
      packages=find_packages(),
      author='Michael Hirsch, Ph.D',
      description='Model of Earth atmosphere absorption and transmission vs. wavelength and location on Earth.',
      long_description=open('README.rst').read(),
      version='2.3.1',
      url='https://github.com/scivision/lowtran',
      classifiers=[
      'Development Status :: 4 - Beta',
      'Environment :: Console',
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: MIT License',
      'Operating System :: OS Independent',
      'Programming Language :: Python :: 3',
      'Topic :: Scientific/Engineering :: Atmospheric Science',
      ],
      ext_modules=ext,
      install_requires=install_requires,
      python_requires='>=3.6',
      extras_require={'plot':['matplotlib','seaborn'],
                      'tests':tests_require},
      tests_require=tests_require,
      scripts=['ScatterRadiance.py','SolarIrradiance.py','ThermalRadiance.py',
               'TransmittanceGround2Space.py','UserDataHorizontalRadiance.py',
               'HorizontalTransmittance.py','Wavelength2LowtranWavenumber.py',
               ],
      include_package_data=True,
	  )



#!/usr/bin/env python
"""
convert wavelength (nm) to wavenumber (cm^-1) with step size for input to Lowtran
"""
from lowtran import nm2lt7
#
from argparse import ArgumentParser
p = ArgumentParser()
p.add_argument('wlnm',help='wavelength [nm]',nargs='+',type=float)
p = p.parse_args()

print(nm2lt7(p.wlnm)[0])



#!/usr/bin/env python
"""
Total Radiance = atmosphere rad. or boundary rad. + atm. scat. or boundary refl.

Lowtran outputs W cm^-2 ster^-1 micron^-1
we want photons cm^-2 s^-1 ster^-1 micron^-1
1 W cm^-2 = 10000 W m^-2

h = 6.62607004e-34 m^2 kg s^-1
I: irradiance
Np: numer of photons
Np = (Ilowtran*10000)*lambda_m/(h*c)
"""
from pathlib import Path
from matplotlib.pyplot import show
try:
    import seaborn as sns
    sns.set_context('talk',font_scale=1.5)
except ImportError:
    pass
#
import lowtran
from lowtran.plots import plotscatter


if __name__=='__main__':
    from argparse import ArgumentParser
    p = ArgumentParser(description='Lowtran 7 interface')
    p.add_argument('-z','--obsalt',help='altitude of observer [km]',type=float,default=0.)
    p.add_argument('-a','--zenang',help='Observer zenith angle [deg] ',nargs='+',type=float,default=[0.,60,80])
    p.add_argument('-w','--wavelen',help='wavelength range nm (start,stop)',type=float,nargs=2,default=(300,1000))
    p.add_argument('-o','--outfn',help='NetCDF4 file to write')
    p.add_argument('--model',help='0-6, see Card1 "model" reference. 5=subarctic winter',type=int,default=5)

    p=p.parse_args()

    c1={'model':p.model,
        'h1': p.obsalt, # of observer
        'angle': p.zenang, # of observer
        'wlnmlim': p.wavelen,
        }
# %%
    TR = lowtran.scatter(c1)
# %%
    if p.outfn:
        outfn = Path(p.outfn).expanduser()
        print('writing', outfn)
        TR.to_netcdf(outfn)
# %%
    plotscatter(TR, c1)

    show()
